using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using N2.Engine;

namespace N2.Web.Drawing
{
	/// <summary>
	/// Parameters for the <see cref="ImageResizer"/>.
	/// </summary>
	public class ImageResizeParameters
	{
		public ImageResizeParameters(double maxWidth, double maxHeight)
			: this(maxWidth, maxHeight, ImageResizeMode.Fit)
		{
		}

		public ImageResizeParameters(double maxWidth, double maxHeight, ImageResizeMode mode)
		{
			Mode = mode;
			MaxWidth = maxWidth;
			MaxHeight = maxHeight;
			Quality = 90;
		}

		public ImageResizeMode Mode { get; set; }
		public double MaxWidth { get; set; }
		public double MaxHeight { get; set; }
		public int Quality { get; set; }

		public Rectangle? SourceRectangle { get; set; }

		public static ImageResizeParameters Fill(Rectangle sourceRectangle, int maxWidth, int maxHeight)
		{
			return new ImageResizeParameters(maxWidth, maxHeight, ImageResizeMode.Fill) { SourceRectangle = sourceRectangle };
		}
	}

	/// <summary>
	/// Resizes an image.
	/// </summary>
	[Service]
    public class ImageResizer
    {
	    [Obsolete("Use overload with parameters")]
		public virtual void Resize(Stream inputStream, string extension, double maxWidth, double maxHeight, Stream outputStream)
		{
			Resize(inputStream, new ImageResizeParameters(maxWidth, maxHeight, ImageResizeMode.Fit), outputStream);
		}

		[Obsolete("Use overload with parameters")]
		public virtual bool Resize(Stream inputStream, string extension, double maxWidth, double maxHeight, ImageResizeMode mode, Stream outputStream)
		{
			return Resize(inputStream, new ImageResizeParameters(maxWidth, maxHeight, mode), outputStream);
		}

		[Obsolete("Use overload with parameters")]
		public virtual byte[] GetResizedBytes(Stream imageStream, string extension, double maxWidth, double maxHeight, ImageResizeMode mode)
		{
			return GetResizedBytes(imageStream, new ImageResizeParameters(maxWidth, maxHeight, mode));
		}

		public virtual bool Resize(Stream inputStream, ImageResizeParameters parameters, Stream outputStream)
		{
            if (inputStream == null) throw new ArgumentNullException("inputStream");

			using (Bitmap original = new Bitmap(inputStream))
			{
				Resize(original, parameters, outputStream);
				return true;
			}
		}

		public virtual byte[] GetResizedBytes(Stream imageStream, ImageResizeParameters parameters)
		{
            if (imageStream == null) throw new ArgumentNullException("imageStream");

			using (Bitmap original = new Bitmap(imageStream))
			{
				var ms = new MemoryStream();
				Resize(original, parameters, ms);
				return ms.GetBuffer();
			}
		}

	    private void Resize(Bitmap original, ImageResizeParameters parameters, Stream output)
    	{
    		Bitmap resized;
			var mode = parameters.Mode;
			var maxWidth = parameters.MaxWidth;
			var maxHeight = parameters.MaxHeight;
			var quality = parameters.Quality;
			var srcRect = parameters.SourceRectangle;

			if (mode == ImageResizeMode.Fit)
			{
				double resizeRatio = GetResizeRatio(original, maxWidth, maxHeight);
				int newWidth = (int)Math.Round(original.Width * resizeRatio);
				int newHeight = (int)Math.Round(original.Height * resizeRatio);
				resized = new Bitmap(newWidth, newHeight, original.PixelFormat);
			}
			else
				resized = new Bitmap((int)maxWidth, (int)maxHeight, original.PixelFormat);

    		resized.SetResolution(original.HorizontalResolution, original.VerticalResolution);

    		Graphics g = CreateGraphics(original, ref resized);

    		using (g)
    		{
    			g.PageUnit = GraphicsUnit.Pixel;
    			g.InterpolationMode = InterpolationMode.HighQualityBicubic;
    			g.SmoothingMode = SmoothingMode.HighQuality;
    			g.PixelOffsetMode = PixelOffsetMode.HighQuality;
    			g.CompositingMode = CompositingMode.SourceCopy;
    			g.CompositingQuality = CompositingQuality.HighQuality;

    			using (ImageAttributes attr = new ImageAttributes())
    			{
    				attr.SetWrapMode(WrapMode.TileFlipXY);

					var originalSize = srcRect.HasValue 
						? srcRect.Value.Size 
						: original.Size;

					Rectangle dest = (mode == ImageResizeMode.Fill)
						? GetFillDestinationRectangle(originalSize, resized.Size)
						: new Rectangle(Point.Empty, resized.Size);
					
					var src = new Rectangle(0, 0, original.Width, original.Height);
					if (srcRect.HasValue)
						src = srcRect.Value;

					g.DrawImage(original, dest, src.X, src.Y, src.Width, src.Height, GraphicsUnit.Pixel, attr);
				}

                // Use higher quality compression if the original image is jpg. Default is 75L.
                ImageCodecInfo codec = GetEncoderInfo(original.RawFormat.Guid);

                if (codec != null && codec.MimeType.Equals("image/jpeg"))
                {
                    EncoderParameters encoderParams = new EncoderParameters(1);
					encoderParams.Param[0] = new EncoderParameter(Encoder.Quality, (long)quality);

                    resized.Save(output, codec, encoderParams);                    
                }
                else
                {
    			    resized.Save(output, original.RawFormat);
                }
    		}
    	}

        private ImageCodecInfo GetEncoderInfo(Guid formatID)
        {
            // Get image codecs for all image formats
            ImageCodecInfo[] codecs = ImageCodecInfo.GetImageEncoders();

            // Find the correct image codec
            for (int i = 0; i < codecs.Length; i++)
                if (codecs[i].FormatID == formatID)
                    return codecs[i];
            return null;
        }

		private Graphics CreateGraphics(Bitmap original, ref Bitmap resized)
		{
			ImageCodecInfo info = GetEncoderInfo(original.RawFormat.Guid);

			switch (info.MimeType)
			{
				case "image/jpeg":
					if (resized.PixelFormat == PixelFormat.Format1bppIndexed || resized.PixelFormat == PixelFormat.Format4bppIndexed || resized.PixelFormat == PixelFormat.Format8bppIndexed)
						return GetResizedBitmap(ref resized, PixelFormat.Format24bppRgb);
					return Graphics.FromImage(resized);
				case "image/gif":
					return GetResizedBitmap(ref resized, PixelFormat.Format24bppRgb);
				case "image/png":
				default:
					return GetResizedBitmap(ref resized, original.PixelFormat);
			}
		}

		private Graphics GetResizedBitmap(ref Bitmap resized, PixelFormat format)
		{
			Bitmap temp = new Bitmap(resized.Width, resized.Height, format);
			var g = Graphics.FromImage(temp);
			g.DrawImage(resized, new Rectangle(0, 0, temp.Width, temp.Height), 0, 0, resized.Width, resized.Height, GraphicsUnit.Pixel);
			resized = temp;
			return g;
		}

		public static Rectangle GetFillDestinationRectangle(Size original, Size resized)
		{
			double resizeWidth = resized.Width / (double)original.Width;
			double resizeHeight = resized.Height / (double)original.Height;
			double resizeMax = Math.Max(resizeWidth, resizeHeight);

			var size = new Size((int)(original.Width * resizeMax), (int)(original.Height * resizeMax));

			bool isBeeingMadeNarrower = resizeWidth < resizeHeight;

			Point relocated = isBeeingMadeNarrower
				? new Point((resized.Width - size.Width) / 2, 0)
				: new Point(0, (resized.Height - size.Height) / 2);

			return new Rectangle(relocated, size);
		}

		double GetResizeRatio(Bitmap original, double width, double height)
		{
			double ratioY = height / original.Height;
			double ratioX = width / original.Width;

			double ratio = Math.Min(ratioX, ratioY);
			if (ratio == 0)
				ratio = Math.Max(ratioX, ratioY);
			if (ratio <= 0 || ratio > 1) ratio = 1;
			return ratio;
		}

		protected virtual void TransferBetweenStreams(Stream inputStream, Stream outputStream)
		{
			byte[] buffer = new byte[32768];
			while (true)
			{
				int bytesRead = inputStream.Read(buffer, 0, buffer.Length);
				if (bytesRead <= 0)
					break;

				outputStream.Write(buffer, 0, bytesRead);
			}
		}
    }
}